/**
 * CarAPI.java
 */

import java.util.Timer;
import java.util.TimerTask;
/**
 * Copyright (c) 2003 Sony Ericsson Mobile Communications AB<br>
 * All Rights Reserved.
 * <p>
 * CarAPI is the interface between the GUI and the other classes.
 * It provides methods for GUI to control the Bluetooth car and 
 * check its status. CarConnection in the other hand retrieves the control 
 * commands from CarAPI and send them to the car.
 *
 * @author	Christopher Wong
 * @author	ke Rydgren
 * @version	2.0
 */
public class CarAPI {

	/**
	 * Constant for car ready to drive.
	 */
	public static final int CAR_READY = 4;
	
	/**
	 * Constant for car charging.
	 */
	public static final int CAR_CHARGING = 5;
	
	/**
	 * Constant for no connection to the car.
	 */
	public static final int CAR_NO_LINK = 6;
	
	/**
	 * The status of the car.
	 */
	private int status;

	/**
	 * Constant for car driving forward with speed 3.
	 */
	public static final int CAR_SPEED_3 = 3;
	
	/**
	 * Constant for car driving forward with speed 2.
	 */
	public static final int CAR_SPEED_2 = 2;
	
	/**
	 * Constant for car driving forward with speed 1.
	 */
	public static final int CAR_SPEED_1 = 1;
	
	/**
	 * Constant for car stopped.
	 */
	public static final int CAR_SPEED_N = 0;
	
	/**
	 * Constant for car driving backward.
	 */
	public static final int CAR_SPEED_R = -1;
	
	/**
	 * The speed of the car.
	 */
	private int speed;
	
	/**
	 * Flag indicate that the car speed has changed.
	 */
	private boolean speedChanged;
	
	/**
	 * Constant for car turning left.
	 */
	public static final int CAR_DIRECTION_LEFT = 7;
	
	/**
	 * Constant for car turning left lightly.
	 */
	public static final int CAR_DIRECTION_LIGHT_LEFT = 8;
	
	/**
	 * Constant for car going straight.
	 */
	public static final int CAR_DIRECTION_STRAIGHT = 9;
	
	/**
	 * Constant for car turning right lightly.
	 */
	public static final int CAR_DIRECTION_LIGHT_RIGHT = 10;
	
	/**
	 * Constant for car turning right.
	 */
	public static final int CAR_DIRECTION_RIGHT = 11;
	
	/**
	 * The direction of the car.
	 */
	private int direction;
	
	/**
	 * Flag indicate that the car direction has changed.
	 */
	private boolean directionChanged;
	
	/**
	 * Flag indicating that the MIDlet is active (i.e. visible and not in Paused state).
	 */
	private boolean appActive = false;
	
	/**
	 * Reference to the BluetoothCar class (the MIDlet).
	 */
	private BluetoothCar bluetoothCar;
	
	/**
	 * Flag indicating that the car connection should close as soon as possible.
	 */
	public boolean timeToCloseConnection = false;
	
	/*
	 * The delay time for the disconnectTimer in milliseconds.
	 */
	private final long TIMER_DELAY = 10000;
	
	/*
	 * The TimerTask that is responsible for setting the timeToCloseConnection
	 * flag to true.
	 */
	private TimerTask disconnectTimerTask;
	
	/* The Timer used for setting the timeToCloseConnection flag to true.
	 */
	private Timer disconnectTimer;
	
	/** 
     * Constructs a new <code>CarAPI</code> object.
     */
    public CarAPI(BluetoothCar bluetoothCar) {
    	// Initialize default values
    	status = CAR_NO_LINK;
    	speed = CAR_SPEED_N;
    	direction = CAR_DIRECTION_STRAIGHT;
    	this.bluetoothCar = bluetoothCar;    	
    }

	/**
	 * Check if the <code>MIDlet</code> is in <i>Active</i> state and shown.
	 *
	 * @return 	<code>true</code> if the <code>MIDlet</code> is <i>Active</i>,
	 *			<code>false</code> otherwise.
	 */
	public boolean isAppActive() {
		return appActive;
	}

	/**
	 * Called when the <code>MIDlet</code> switches between the <i>Paused</i>
	 * and the <i>Active</i> state and when the <code>Canvas</code> in
	 * <code>GUI</code> switches between being hidden and shown.
	 * The <code>GUI Canvas</code> becomes hidden when it's partly or completely
	 * obscured by another application window.
	 *
	 * If we are entering the <i>Paused</i> state or being hidden, the car is
	 * stopped and the <code>disconnectTimer</code> is started.
	 *
	 * If we are entering the <i>Active</i> state or being shown, any thread
	 * trapped in the <code>waitUntilActive()</code> method will be signalled
	 * to exit from its <i>wait</i> state and the <code>disconnectTimer</code>
	 * is stopped.
	 *
	 * @param active	<code>true</code> if we are entering the <i>Active</i> 
	 *					state or being shown, <code>false</code> if we are
	 *					entering the <i>Paused</i> state or being hidden.
	 */
	public void setAppActive(boolean active) {
		appActive = active;
		if (active) {
			stopTimer();
			stopWaiting();
		}
		else {
			setSpeed(CAR_SPEED_N);
			setDirection(CAR_DIRECTION_STRAIGHT);
			startDisconnectTimer();
		}
	}

	/**
	 * Starts a timer that will set the flag timeToCloseConnection, which is 
	 * checked by the processEvents method in CarConnection (looped from SPPServer).
	 * The SPPServer will close the connection when this flag is set.
	 */
	public void startDisconnectTimer() {
		stopTimer();
		if (disconnectTimer == null) {
		  disconnectTimer = new Timer();
       	  // Create the disconnectTimerTask.
	      disconnectTimerTask = new TimerTask() {
	          public void run() {
	          	  if (!isAppActive()) timeToCloseConnection = true;
	          }
	      };
		  disconnectTimer.schedule(disconnectTimerTask, TIMER_DELAY); 
		}
	}

	/**
	 * Stops the disconnectTimer.
	 */
	public void stopTimer() {
	    if (disconnectTimer != null) {
	      disconnectTimer.cancel();
	    }
        disconnectTimer = null;
	}
	
	/**
	* Puts the calling thread into wait mode if the MIDlet is inactive or not
	* visible and if there is no connection to the car and the GUI has processed
	* that state (i.e. the CAR_NO_LINK state).
	*/
	public synchronized void waitUntilActive() {
		if (!appActive && status == CAR_NO_LINK && bluetoothCar.guiUpdated())
		try {
			wait();
		} catch (InterruptedException ie) {}
	}
	
	/**
	 * Releases any thread that is trapped in wait mode in the waitUntilActive()
	 * method above.
	 */
	public synchronized void stopWaiting() {
		notifyAll();
	}

	/**
	 * Sets the current status of the Bluetooth car.
	 * If the GUI is trapped waiting for the application to become active, and 
	 * the new status is different from the previous, the GUI is released from
	 * the wait() method and starts running again.
	 *
	 * @param s	the status to be set. It has to be either 
	 *			<code>CAR_READY</code>, <code>CAR_CHARGING</code>, or 
	 *			<code>CAR_NO_LINK</code>.
	 */
	public void setStatus(int s) {
		// No need to set value if it is the same as the current value.
		if (s == status)
			return;
			
		// Make sure that only valid values are set.
		if (s >= CAR_READY && s <= CAR_NO_LINK) {
			status = s;
			stopWaiting();
		}
	}

	/**
	 * Returns the current status of the Bluetooth car.
	 *
	 * @return the current status fo the Bluetooth car
	 */
	public int getStatus() {
		return status;	
	}

	/**
	 * Check if a new speed is set for the Bluetooth car.
	 *
	 * @return 	<code>true</code> if the car speed has changed<br>  
	 *			<code>false</code> otherwise
	 */
	public boolean isSpeedChanged() {
		return speedChanged;
	}

	/**
	 * Controls the speed of the car. 
	 *
	 * @param s	this can be <code>CAR_SPEED_R</code>, 
	 *			<code>CAR_SPEED_N</code>, <code>CAR_SPEED_1</code>,
	 *			<code>CAR_SPEED_2</code>, or <code>CAR_SPEED_3</code>.
	 */
	public void setSpeed(int s) {
		// No need to set value if it is the same as the current value.
		if (s == speed)
			return;
			
		// Make sure that only valid values are set.
		if (s >= CAR_SPEED_R && s <= CAR_SPEED_3) {
			speed = s;	
			speedChanged = true;
		}
	}

	/**
	 * Increase the speed by one.
	 */
	public void incSpeed() {
		// If speed is already as high as possible.
		if (speed == CAR_SPEED_3)
			return;
		
		// Else increase speed by one.
		speed++;
		speedChanged = true;
	}

	/**
	 * Decrease the speed by one.
	 */
	public void decSpeed() {
		// If speed is already as low as possible.
		if (speed == CAR_SPEED_R)
			return;
		
		// Else decrease speed by one.
		speed--;
		speedChanged = true;
	}

	/**
	 * Retrieves the speed set for the Bluetooth car.
	 *
	 * @return 	the speed set. It can be either <code>CAR_SPEED_R</code>, 
	 *			<code>CAR_SPEED_N</code>, <code>CAR_SPEED_1</code>,
	 *			<code>CAR_SPEED_2</code>, or <code>CAR_SPEED_3</code>.
	 */
	public int getSpeed() {
		return speed;	
	}

	/**
	 * Clears the flag and retrieves the speed set for the Bluetooth car.
	 *
	 * @return 	the speed set. It can be either <code>CAR_SPEED_R</code>, 
	 *			<code>CAR_SPEED_N</code>, <code>CAR_SPEED_1</code>,
	 *			<code>CAR_SPEED_2</code>, or <code>CAR_SPEED_3</code>.
	 */
	public int getAndClearSpeed() {
		speedChanged = false;
		return speed;	
	}

	/**
	 * Check if a new direction is set for the Bluetooth car.
	 *
	 * @return 	<code>true</code> if the car direction has changed<br>  
	 *			<code>false</code> otherwise
	 */
	public boolean isDirectionChanged() {
		return directionChanged;
	}

	/**
	 * Controls the speed of the car. 
	 *
	 * @param d	this can be <code>CAR_DIRECTION_LEFT</code>, 
	 *			<code>CAR_DIRECTION_LIGHT_LEFT</code>, 
	 *			<code>CAR_DIRECTION_STRAIGHT</code>,
	 *			<code>CAR_DIRECTION_LIGHT_RIGHT</code>, or 
	 *			<code>CAR_DIRECTION_RIGHT</code>.
	 */
	public void setDirection(int d) {
		// No need to set value if it is the same as the current value.
		if (d == direction)
			return;
			
		// Make sure that only valid values are set.
		if (d >= CAR_DIRECTION_LEFT && d <= CAR_DIRECTION_RIGHT) {
			direction = d;	
			directionChanged = true;
		}
	}

	/**
	 * Sets the car direction to turn one step to the left.
	 */
	public void turnLeft() {
		// If direction is already as long to the left as possible.
		if (direction == CAR_DIRECTION_LEFT)
			return;
		
		// Else one step to the left.
		direction--;
		directionChanged = true;
	}

	/**
	 * Sets the car direction to turn one step to the right.
	 */
	public void turnRight() {
		// If direction is already as long to the right as possible.
		if (direction == CAR_DIRECTION_RIGHT)
			return;
		
		// Else one step to the right.
		direction++;
		directionChanged = true;
	}

	/**
	 * Retrieves the direction set for the Bluetooth car.
	 *
	 * @return 	the direction set. It can be either 
	 *			<code>CAR_DIRECTION_LEFT</code>, 
	 *			<code>CAR_DIRECTION_LIGHT_LEFT</code>, 
	 *			<code>CAR_DIRECTION_STRAIGHT</code>,
	 *			<code>CAR_DIRECTION_LIGHT_RIGHT</code>, or 
	 *			<code>CAR_DIRECTION_RIGHT</code>.
	 */
	public int getDirection() {
		return direction;	
	}

	/**
	 * Clears the flag and retrieves the direction set for the Bluetooth car.
	 *
	 * @return 	the direction set. It can be either 
	 *			<code>CAR_DIRECTION_LEFT</code>, 
	 *			<code>CAR_DIRECTION_LIGHT_LEFT</code>, 
	 *			<code>CAR_DIRECTION_STRAIGHT</code>,
	 *			<code>CAR_DIRECTION_LIGHT_RIGHT</code>, or 
	 *			<code>CAR_DIRECTION_RIGHT</code>.
	 */
	public int getAndClearDirection() {
		directionChanged = false;
		return direction;	
	}

	/**
	 * This will stop the car.
	 */
	public void stop() {
		// No need to set value if it is the same as the current value.
		if (speed == CAR_SPEED_N)
			return;
			
		speed = CAR_SPEED_N;
		speedChanged = true;
	}

}
