/**
 * CarConnection.java 
 */

import java.io.InputStream;
import java.io.OutputStream;
import java.io.IOException;
import javax.microedition.io.*;

/**
 * Copyright (c) 2003 Sony Ericsson Mobile Communications AB<br>
 * All Rights Reserved.
 * <p>
 * CarConnection is the interface between the car and the MIDlet. All car commands is stored
 * in this class. This object will retrieve user commands and send them to the
 * car. It will also set the car status, such as when the car is not connected,
 * charging or ready to run.
 * 
 *
 * @author	Christopher Wong
 * @author	ke Rydgren
 * @version	2.1
 */
public class CarConnection {
	// The string which says that the car is charging.
	private final String CAR_CHARGING 	= "Charging";
	// The string which says that the car is ready for driving.
	private final String CAR_SPEED 		= "Speed=2";
	// The string which says that the car is ready for driving.
	private final String CAR_READY 		= "Ready";
	
	// Car control commands.
	private final String FORWARD_PRESSED 	= "+CKEV: u,1";
	private final String BACK_PRESSED 		= "+CKEV: d,1";
	private final String SPEED3_PRESSED 	= "+CKEV: #,1";
	private final String SPEED2_PRESSED 	= "+CKEV: 0,1";
	private final String SPEED1_PRESSED 	= "+CKEV: *,1";
	private final String STOP_PRESSED 		= "+CKEV: :J,1";
	private final String LEFT_PRESSED 		= "+CKEV: 2,1";
	private final String LEFT_RELEASED 		= "+CKEV: 2,0";
	private final String RIGHT_PRESSED 		= "+CKEV: 3,1";
	private final String RIGHT_RELEASED		= "+CKEV: 3,0";
	private final String NO_PRESSED			= "+CKEV: e,1";
	private final String NO_RELEASED		= "+CKEV: e,0";
	
	// The response to the car.
	private final String OK = "OK";
	
	// Init carriage return and line feed.
	private final byte[] CRLF = {13, 10};
	
	// Available gear values.
	private final int GEAR_3 = 3;
	private final int GEAR_2 = 2;
	private final int GEAR_1 = 1;
	// The current gear in the car.
	private int currGear;
	
	// Available acceleration values.
	private final int ACC_ON 	= 4;
	private final int ACC_OFF 	= 5;
	private final int ACC_BACK 	= 6;
	// The current acceleration in the car.
	private int currAcceleration;
	
	// The current direction in the car. It uses the same values as the CarAPI.
	private int currDirection;
	
	// Used to control car and check car status.
	private CarAPI carAPI;

	// Delay between every user command handling. This is used to avoid building
	// up a queue of commands in the car.
	private final long SEND_DELAY = 100;

	// This time is set after each data send.
	private long lastSendTime;
	
	/**
	 * The constructor for creating a <code>CarConnection</code> object,
	 * which sends command to the car and listen for data from the car.
	 * <code>CarAPI</code> is needed to be able to retrieve user input.
	 *
	 * @param carAPI	the <code>CarAPI</code> object needed to retrieve user
	 *					input
	 */
    public CarConnection(CarAPI carAPI) {
    	this.carAPI = carAPI;
		lastSendTime = System.currentTimeMillis();
		resetCarData();
    }
    
	/**
	 * Resets the internal car data to the state it's supposed to be in when it
	 * first connects to the phone.
	 */
	public void resetCarData() {
		currGear = GEAR_2;
		currAcceleration = ACC_OFF;
		currDirection = carAPI.CAR_DIRECTION_STRAIGHT;
	}
	
	/**
	 * Tell the car that we want to quit by sending the NO keypress.
	 */
	public void quitCarLink(OutputStream os) {
		try {
			send(os, NO_PRESSED);
			send(os, NO_RELEASED);
		} catch (IOException e) {}
		// Set car status.
		carAPI.setStatus(carAPI.CAR_NO_LINK);
	}
	
	/**
	 * Processes all messages from the car (AT-commands) and all user events
	 * from the GUI (via the CarAPI).
	 *
	 * @param is The <code>InputStream</code> from the car connection
	 *
	 * @param os The <code>OutputStream</code> to the car connection
	 */
	public boolean processEvents(InputStream is, OutputStream os) {
		try {
			// If there is anything in the input stream, it got the 
			// highest priority.
			int bytesToRead = is.available();
			if (bytesToRead > 0) {
				// Initialize buffer
				byte[] byteBuffer = new byte[bytesToRead];
				// Read bytes
				int nbrOfBytesRead = is.read(byteBuffer);
			    String str = new String(byteBuffer);
			    // Print the received message.
			    System.out.println("RX: " + str);
			    
			    // Set the car status.
			    if (str.indexOf(CAR_READY) != -1 ||
			    	str.indexOf(CAR_SPEED) != -1)
			    	carAPI.setStatus(carAPI.CAR_READY);
			  	else if (str.indexOf(CAR_CHARGING) != -1)
			    	carAPI.setStatus(carAPI.CAR_CHARGING);
			    	
			    // Answer with OK.
			    sendOK(os);
			}
		
		
			// Take care of commands issued by the user.
			else if (System.currentTimeMillis() - lastSendTime > SEND_DELAY &&
					 carAPI.getStatus() == carAPI.CAR_READY) {
	
				// Handle speed change
				if (carAPI.isSpeedChanged()) {
					
					int speed = carAPI.getAndClearSpeed();
					
					if (speed == carAPI.CAR_SPEED_3 || 
						speed == carAPI.CAR_SPEED_2 ||
						speed == carAPI.CAR_SPEED_1) {
						
						// Set gear.
						if (speed == carAPI.CAR_SPEED_3 && 
							currGear != GEAR_3) {
							send(os, SPEED3_PRESSED);
							currGear = GEAR_3;	
						}
						else if (speed == carAPI.CAR_SPEED_2 && 
							currGear != GEAR_2) {
							send(os, SPEED2_PRESSED);
							currGear = GEAR_2;	
						}
						else if (speed == carAPI.CAR_SPEED_1 && 
							currGear != GEAR_1) {
							send(os, SPEED1_PRESSED);
							currGear = GEAR_1;	
						}
					
						// Set acceleration.
						if (currAcceleration != ACC_ON) {
							send(os, FORWARD_PRESSED);
							currAcceleration = ACC_ON;
						}
					}
					else if (speed == carAPI.CAR_SPEED_N) {
						// Set acceleration.
						if (currAcceleration != ACC_OFF) {
							send(os, STOP_PRESSED);
							currAcceleration = ACC_OFF;
						}
					}
					else if (speed == carAPI.CAR_SPEED_R) {
						// Set acceleration.
						if (currAcceleration != ACC_BACK) {
							send(os, BACK_PRESSED);
							currAcceleration = ACC_BACK;
						}	
					}
				} // End of speed change
				
				
					
				// Handle direction change
				int direction = carAPI.getAndClearDirection();
				
				// If acceleration is off, set direction to straight to
				// save battery. 
				if (currAcceleration == ACC_OFF) {
					if (currDirection == carAPI.CAR_DIRECTION_LEFT)
						send(os, LEFT_RELEASED);
					else if (currDirection == carAPI.CAR_DIRECTION_RIGHT)
						send(os, RIGHT_RELEASED);
				}	
				else if (direction == carAPI.CAR_DIRECTION_LEFT ||
					(direction == carAPI.CAR_DIRECTION_LIGHT_LEFT &&
					 currDirection != carAPI.CAR_DIRECTION_LEFT)) {
					// Turn left
					send(os, LEFT_PRESSED);
					if (currDirection == carAPI.CAR_DIRECTION_RIGHT)
						send(os, RIGHT_RELEASED);	
					currDirection = carAPI.CAR_DIRECTION_LEFT;
				}
				else if (direction == carAPI.CAR_DIRECTION_RIGHT ||
					(direction == carAPI.CAR_DIRECTION_LIGHT_RIGHT &&
					 currDirection != carAPI.CAR_DIRECTION_RIGHT)) {
					// Turn right
					send(os, RIGHT_PRESSED);
					if (currDirection == carAPI.CAR_DIRECTION_LEFT)
						send(os, LEFT_RELEASED);	
					currDirection = carAPI.CAR_DIRECTION_RIGHT;
				}
				else if (direction == carAPI.CAR_DIRECTION_STRAIGHT || 
						(currDirection != carAPI.CAR_DIRECTION_STRAIGHT &&
						 (direction == carAPI.CAR_DIRECTION_LIGHT_LEFT ||
						  direction == carAPI.CAR_DIRECTION_LIGHT_RIGHT))) {
					// Turn straight
					if (currDirection == carAPI.CAR_DIRECTION_RIGHT)
						send(os, RIGHT_RELEASED);	
					else if (currDirection == carAPI.CAR_DIRECTION_LEFT)
						send(os, LEFT_RELEASED);
					currDirection = carAPI.CAR_DIRECTION_STRAIGHT;
				}
				// End of direction change
				
				
			} // End of commands issued by user.
			
			if (carAPI.timeToCloseConnection) {
				quitCarLink(os);
				return false;
			}
			else {
				return true;
			}
		}
		catch (IllegalArgumentException  e) {
			System.out.println("IllegalArgumentException: " + e.getMessage());
			carAPI.setStatus(carAPI.CAR_NO_LINK);
			return false;
		}
		catch (IOException e) {
	    	System.out.println("IOException: " + e.getMessage());
			carAPI.setStatus(carAPI.CAR_NO_LINK);
			return false;
		}
	} // End of run()	

	/**
	 * Write and flush the specified string to the output stream.
	 *
	 * @param data the string to be sent to output stream
	 */
	private void send(OutputStream os, String data) throws IOException {
		if (os != null && data != null) {		
			os.write(CRLF);	
			os.write(data.getBytes());
			os.write(CRLF);
			os.flush();
			//System.out.println("TX: " + data);
			lastSendTime = System.currentTimeMillis();
		}
	}

	/**
	 * Write and flush an OK string to the output stream.
	 */
	private void sendOK(OutputStream os) throws IOException {
		os.write(CRLF);	
		os.write(OK.getBytes());
		os.write(CRLF);
		os.flush();
		//System.out.println("TX: " + OK);
		lastSendTime = System.currentTimeMillis();
	}
}
