/**
 * GUI.java
 */

import javax.microedition.midlet.*;
import javax.microedition.lcdui.*;
import javax.microedition.lcdui.Display;
import javax.microedition.lcdui.game.*;
import javax.microedition.media.*;
import java.io.*;
import java.util.*;

 
/**
 * Copyright (c) 2003 Sony Ericsson Mobile Communications AB<br>
 * All Rights Reserved.
 * <p>
 * The GUI is the canvas the user sees. It provides graphics & sound
 * for the user and handles user input on the screen or by buttons.
 *
 * @author	Christopher Wong
 * @author	ke Rydgren
 * @version	1.13
 */
public class GUI extends Canvas implements Runnable, CommandListener {

	/**
	 * Constants for P900 keys.
	 */
	private static final int P900_OK = 63562;
	private static final int P900_RET = 63557;
	private static final int P900_C = 63564;
	private static final int P900_MENU = 63563;

	private static final int DISPLAY_IDLE = -2;
	private static final int DISPLAY_REC = -3;
	private static final int DISPLAY_PLAY = -4;

	// The about command
    private Command aboutCommand; 	
    
	// The rotate screen command
    private Command rotateScrCommand; 	
    
	// The exit command
    private Command exitCommand; 	
    
	// The sound FX commands
    private Command soundOnCommand; 	
    private Command soundOffCommand; 	
    
	// The Record/Replay commands
    private Command recordCommand; 	
    private Command stopRecordCommand; 	
    private Command replayCommand; 	
    private Command stopReplayCommand; 	
    
	// Contains the picture for the about box.
	private Image aboutImage;
	
	// Contains the picture for the idle "display".
	private Sprite spDisplayIdle;
	
	// Contains the picture for the recording "display".
	private Sprite spDisplayRec;
	
	// Contains the picture for the replaying "display".
	private Sprite spDisplayPlay;
	
	// Contains the pictures for the wheel, speed slider, and panel.
	private Sprite[] sprites;
	
	// All the pictures used in the application:
	private Image pictSpeedR;
	private Image pictSpeedN;
	private Image pictSpeed1;
	private Image pictSpeed2;
	private Image pictSpeed3;
	private Image pictDisplayIdle;
	private Image pictDisplayRec;
	private Image pictDisplayPlay;
	private Image pictPanelNC;
	private Image pictPanelCh;
	private Image pictPanelOk;
	private Image pictWheelLeft2;
	private Image pictWheelLeft1;
	private Image pictWheelStraight;
	private Image pictWheelRight1;
	private Image pictWheelRight2;
	
	private int lastKnownWidth;
	private int lastKnownHeight;
	private boolean blackScreen;
	private String blackScreenText;

	// Used to control car and check car status.
	private CarAPI carAPI;
	
	// This thread.
	private Thread myThread;
	
	// Full-screen mode if flag is true
	private boolean landscapeMode;
	
	private boolean wheelSoundAvailable;
	private boolean engineSoundAvailable;
	private boolean gearSoundAvailable;
	private boolean statusSoundAvailable;
	private boolean enableWheelSound;
	private boolean enableEngineSound;
	private boolean enableGearSound;
	private boolean enableStatusSound;
	
	private boolean wantToPlayGearUp;
	private boolean wantToPlayGearDown;
	private boolean wantToPlayWheel;
	private boolean wantToPlayReady;
	private boolean wantToPlayCharging;
	private boolean wantToPlayNoLink;
	
	private boolean enableTouchWheel;
	private int lastKnownDirection;
	private int lastKnownSpeed;
	private int lastKnownStatus;
	private int lastKnownMode;
	
	private long startedWheelSoundTime;
	private Player soundGearUp;
	private Player soundGearDown;
	private Player soundGearUpWheel;
	private Player soundGearDownWheel;
	private Player soundWheel;
	private Player soundSpeedR;
	private Player soundSpeedN;
	private Player soundSpeed1;
	private Player soundSpeed2;
	private Player soundSpeed3;
	private Player soundReady;
	private Player soundCharging;
	private Player soundNoLink;
	private Player currentlyPlayingSound;
	
	// Flag that indicates that the thread is started.
	private boolean started;
	
	// Variables used for recording/playback:
	private long eventStartTime;
	private boolean recording = false;
	private boolean replaying = false;
	private Enumeration events;
	private Vector drivePath = new Vector();
	class driveEventClass {
		public int event;
		public long eventDuration;
	};
	private Timer playbackTimer;
	private TimerTask playbackTimerTask = new TimerTask() {
	        public void run() {
				if (replaying && carAPI.isAppActive() && events.hasMoreElements()) {
					driveEventClass driveEvent = (driveEventClass) events.nextElement();
					if (driveEvent.eventDuration >= 0) {
						// Direction:
						if (driveEvent.event >= carAPI.CAR_DIRECTION_LEFT && driveEvent.event <= carAPI.CAR_DIRECTION_RIGHT) {
							carAPI.setDirection(driveEvent.event);
							processCarEvents();
						}
						// Speed:
						else if (driveEvent.event >= carAPI.CAR_SPEED_R && driveEvent.event <= carAPI.CAR_SPEED_3) {
							carAPI.setSpeed(driveEvent.event);
							processCarEvents();
						}
					    if (playbackTimer != null)
							playbackTimer.schedule(playbackTimerTask, driveEvent.eventDuration); 
					} else
						stopReplaying();
				} else
					stopReplaying();
	        }
	    };
	
		
	/**
	 * Reference to the BluetoothCar class (the MIDlet).
	 */
	private BluetoothCar bluetoothCar;
		
	/**
	 * Constructs a new <code>GUI</code> object.
	 *
	 * @param carAPI the interface to control car and check car status.
	 */
	public GUI(CarAPI carAPI, BluetoothCar bluetoothCar)	{
		super();
		// Loads the image used in the about box:
		this.carAPI = carAPI;
		try	{
		  aboutImage = Image.createImage("/About.png");
		}
		catch (IOException e) {}
		
		started = false;
		blackScreen = false;
		blackScreenText = "Loading...";
		lastKnownWidth = 0;
		lastKnownHeight = 0;
		
		if (getHeight() > getWidth())
			landscapeMode = true;
		else
			landscapeMode = false;

		// Declare how many Sprite the vector sprites can store
		sprites = new Sprite[13];
		
		// Create menu commands:
		// Add About command
		aboutCommand = new Command("About", Command.SCREEN, 1);
        addCommand(aboutCommand);
		// Add Record/Replay command
		recordCommand = new Command("Record", "Record Path", Command.SCREEN, 1);
		stopRecordCommand = new Command("Stop Recording", "Stop Recording Path", Command.STOP, 1);
		replayCommand = new Command("Replay", "Replay Path", Command.SCREEN, 1);
		stopReplayCommand = new Command("Stop Replaying", "Stop Replaying Path", Command.STOP, 1);
        addCommand(recordCommand);
		// Add Rotate Screen command
		rotateScrCommand = new Command("Rotate", "Rotate Screen", Command.SCREEN, 1);
        addCommand(rotateScrCommand);
		// Add Exit command
		exitCommand = new Command("Exit", Command.EXIT, 1);
        addCommand(exitCommand);
        // The sound on/off commands must be added after we've loaded the sound files.
		soundOffCommand = new Command("Sound Off", "Sound FX Off", Command.SCREEN, 1);
		soundOnCommand = new Command("Sound On", "Sound FX On", Command.SCREEN, 1);

        setCommandListener(this);
			
		enableTouchWheel = false;
		lastKnownDirection = carAPI.CAR_DIRECTION_STRAIGHT;
		lastKnownSpeed = carAPI.CAR_SPEED_N;
 		lastKnownStatus = carAPI.CAR_NO_LINK;
 		lastKnownMode = getMode();
 		
        myThread = new Thread(this);
		myThread.start();
	}

	/**
	 * The GUI-loop. This loop updates the GUI when a changed has been made to
	 * the car's status, such as car connected, no link to the car, or car charging.
	 * It also initiates any sound FX. The loop will enter the <i>wait</i> state
	 * while the <code>MIDlet</code> is <i>Paused</i> or hidden.
	 */
	public void run() {
		if (!started) {
			started = true;
			blackScreen = true;
			blackScreenText = "Loading graphics...";
			repaint();
			serviceRepaints();
			loadImages();
			blackScreen = true;
			blackScreenText = "Loading Sound FX...";
			repaint();
			serviceRepaints();
			loadSounds();
			addCommands();
			repaint();
		}
 		// Start game loop
 		long lastActiveTime;
		while (true) {
			
			// Check if anything has changed
			if (!allUpdated()) {
					processCarEvents();
			}
			playSounds();
      		// Sleep a short time to let other thread execute.
      		try {
      			myThread.sleep(50);
      		}
      		catch (InterruptedException e) {
            	System.out.println("InterruptedException: " + e.getMessage());
      		}
      		lastActiveTime = System.currentTimeMillis();
      		// Go into wait mode if the MIDlet is inactive or not visible.
      		carAPI.waitUntilActive();
      		
      	} // End of game loop
      	
	}
	
	public boolean allUpdated() {
		return (carAPI.getStatus() == lastKnownStatus && 
			carAPI.getDirection() == lastKnownDirection && 
			carAPI.getSpeed() == lastKnownSpeed &&
			getMode() == lastKnownMode);
		
	}

	/**
	 * Returns a <code>Player</code> prepared with the sound file specified.
	 *
	 * @param fileName the filename of the sound file
	 * @param loop true if the sound shall be looped
	 * @return the <code>Player</code> to represent the sound FX
	 */ 
	private Player getSoundFX(String fileName, boolean loop) {
		boolean result = true;
		Player player = null;
		try {
		    InputStream is = getClass().getResourceAsStream(fileName);
		    player = Manager.createPlayer(is, "audio/X-wav");
		    player.prefetch();
			if (loop)
				player.setLoopCount(-1);
		} catch (Exception e) { result = false; }
		if (result)
			return player;
		else
			return null;
	}
	
	/**
	 * Loads all images used in the application from the JAR file, downscale any
	 * image as needed, and inserts these images into sprites. Finally the
	 * setSpritePos() is called.
	 */ 
	private void loadImages() {
		if (!started) return;
		// Loads all images (if not already loaded) from the JAR file:
		try	{
		  if (pictSpeedR == null) pictSpeedR = Image.createImage("/SpeedR.png");
		  if (pictSpeedN == null) pictSpeedN = Image.createImage("/SpeedN.png");
		  if (pictSpeed1 == null) pictSpeed1 = Image.createImage("/Speed1.png");
		  if (pictSpeed2 == null) pictSpeed2 = Image.createImage("/Speed2.png");
		  if (pictSpeed3 == null) pictSpeed3 = Image.createImage("/Speed3.png");
		  if (pictDisplayIdle == null) pictDisplayIdle = Image.createImage("/DisplayIdle.png");
		  if (pictDisplayRec == null) pictDisplayRec = Image.createImage("/DisplayRec.png");
		  if (pictDisplayPlay == null) pictDisplayPlay = Image.createImage("/DisplayPlay.png");
		  if (pictPanelNC == null) pictPanelNC = Image.createImage("/NoLink.png");
		  if (pictPanelCh == null) pictPanelCh = Image.createImage("/Charging.png");
		  if (pictPanelOk == null) pictPanelOk = Image.createImage("/Ready.png");
		  if (pictWheelLeft2 == null) pictWheelLeft2 = Image.createImage("/WheelL2.png");
		  if (pictWheelLeft1 == null) pictWheelLeft1 = Image.createImage("/WheelL1.png");
		  if (pictWheelStraight == null) pictWheelStraight = Image.createImage("/Wheel0.png");
		  if (pictWheelRight1 == null) pictWheelRight1 = Image.createImage("/WheelR1.png");
		  if (pictWheelRight2 == null) pictWheelRight2 = Image.createImage("/WheelR2.png");
		}
		catch (java.io.IOException e) {}

		// Calculate the downscaling (if any) of the images. The speed control
		// image is scaled separately from the wheel and panel images, and the 
		// display image is not scaled at all (it would look really bad if scaled).
		int imageScale1;
		int imageScale2;
		int screenWidth;
		int screenHeight;
		int displayPictWidth = pictDisplayIdle.getWidth();
		int displayPictHeight = pictDisplayIdle.getHeight();
		int panelPictHeight = pictPanelOk.getHeight();
		int wheelPictHeight = pictWheelStraight.getHeight();
		int sliderPictHeight = pictSpeedN.getWidth();
		int sliderPictWidth = pictSpeedN.getHeight();
		
		if (landscapeMode) {
			screenWidth = getHeight();
			screenHeight = getWidth();
		} else {
			screenWidth = getWidth();
			screenHeight = getHeight();
		}
		// Sorry for the complicated expression below... 
		boolean preferredLayout =
			(((sliderPictWidth * ((screenHeight - displayPictHeight) * 100))
			/ sliderPictHeight) / 100)
			<= screenWidth - displayPictWidth;
		if (preferredLayout) {
			// The slider in one column to the left and the other images in one column to its right.

			// Scale factor for the Speed slider:
			int originalWidth = sliderPictWidth;
			int originalHeight = sliderPictHeight;
			int relativeScreenWidth100 = (screenWidth - displayPictWidth) * 100 / originalWidth;
			int relativeScreenHeight100 = screenHeight * 100 / originalHeight;
			if (relativeScreenWidth100 < relativeScreenHeight100)
				imageScale1 = relativeScreenWidth100;
			else
				imageScale1 = relativeScreenHeight100;

			// Scale factor for the Wheel & Panel:
			originalWidth = displayPictWidth;
			originalHeight = panelPictHeight + wheelPictHeight;
			relativeScreenWidth100 = (screenWidth - ((sliderPictWidth * imageScale1) / 100)) * 100 / originalWidth;
			relativeScreenHeight100 = (screenHeight - displayPictHeight) * 100 / originalHeight;
			if (relativeScreenWidth100 < relativeScreenHeight100)
				imageScale2 = relativeScreenWidth100;
			else
				imageScale2 = relativeScreenHeight100;
		} else {
			// The Display image is centered at the top, and below it are the
			// Panel & Wheel in one column to the right and the Slider to the left.
			
			// Scale factor for the Speed slider:
			int originalHeight = sliderPictHeight;
			int relativeScreenHeight100 = (screenHeight - displayPictHeight) * 100 / originalHeight;
			imageScale1 = relativeScreenHeight100;

			// Scale factor for the Wheel & Panel:
			int originalWidth = displayPictWidth;
			originalHeight = wheelPictHeight + panelPictHeight;
			int relativeScreenWidth100 = (screenWidth - ((sliderPictWidth * imageScale1) / 100)) * 100 / originalWidth;
			relativeScreenHeight100 = (screenHeight - displayPictHeight) * 100 / originalHeight;
			if (relativeScreenWidth100 < relativeScreenHeight100)
				imageScale2 = relativeScreenWidth100;
			else
				imageScale2 = relativeScreenHeight100;
		}
		
		// Load all images into Sprites.
		try	{
		  	spDisplayIdle = new Sprite(pictDisplayIdle);
		  	spDisplayRec = new Sprite(pictDisplayRec);
		  	spDisplayPlay = new Sprite(pictDisplayPlay);
		  	
		  	insertSprite(carAPI.CAR_NO_LINK, 
		  				 new Sprite(downscaleImage(imageScale2, pictPanelNC)));
		  	insertSprite(carAPI.CAR_CHARGING, 
		  				 new Sprite(downscaleImage(imageScale2, pictPanelCh)));
		  	insertSprite(carAPI.CAR_READY, 
		  				 new Sprite(downscaleImage(imageScale2, pictPanelOk)));
		  	insertSprite(carAPI.CAR_SPEED_R, 
		  				 new Sprite(downscaleImage(imageScale1, pictSpeedR)));
		  	insertSprite(carAPI.CAR_SPEED_N, 
		  				 new Sprite(downscaleImage(imageScale1, pictSpeedN)));
		  	insertSprite(carAPI.CAR_SPEED_1, 
		  				 new Sprite(downscaleImage(imageScale1, pictSpeed1)));
		  	insertSprite(carAPI.CAR_SPEED_2, 
		  				 new Sprite(downscaleImage(imageScale1, pictSpeed2)));
		  	insertSprite(carAPI.CAR_SPEED_3, 
		  				 new Sprite(downscaleImage(imageScale1, pictSpeed3)));
		  	insertSprite(carAPI.CAR_DIRECTION_LEFT, 
		  				 new Sprite(downscaleImage(imageScale2, pictWheelLeft2)));
		  	insertSprite(carAPI.CAR_DIRECTION_LIGHT_LEFT, 
		  				 new Sprite(downscaleImage(imageScale2, pictWheelLeft1)));
		  	insertSprite(carAPI.CAR_DIRECTION_STRAIGHT, 
		  				 new Sprite(downscaleImage(imageScale2, pictWheelStraight)));
		  	insertSprite(carAPI.CAR_DIRECTION_LIGHT_RIGHT, 
		  				 new Sprite(downscaleImage(imageScale2, pictWheelRight1)));
		  	insertSprite(carAPI.CAR_DIRECTION_RIGHT, 
		  				 new Sprite(downscaleImage(imageScale2, pictWheelRight2)));
		}
		catch (NullPointerException e) {
            System.out.println("NullPointerException: Load sprites: " + 
            					e.getMessage());
		}
		
		// Set the position for all sprites.
		setSpritePos(preferredLayout);
		// Remember the current width & height.
		lastKnownWidth = getWidth();
		lastKnownHeight = getHeight();
		repaint();
	}
	
	/**
	 * Loads all sounds used in the application from the JAR file.
	 */ 
	private void loadSounds() {
		if (!started) return;
		gearSoundAvailable = true;
		wheelSoundAvailable = true;
		engineSoundAvailable = true;
		statusSoundAvailable = true;
		currentlyPlayingSound = null;
		startedWheelSoundTime = System.currentTimeMillis();
	
		// loading the sounds:
		statusSoundAvailable = ((soundReady = getSoundFX("Ready.wav", false)) != null &&
								(soundCharging = getSoundFX("Charging.wav", false)) != null &&
								(soundNoLink = getSoundFX("NoLink.wav", false)) != null);
		wheelSoundAvailable = ((soundWheel = getSoundFX("Wheel.wav", false)) != null);
		gearSoundAvailable = ((soundGearUp = getSoundFX("GearUp.wav", false)) != null &&
								(soundGearDown = getSoundFX("GearDown.wav", false)) != null);
		if (gearSoundAvailable && wheelSoundAvailable) {
			soundGearUpWheel = getSoundFX("GearUpWheel.wav", false);
			soundGearDownWheel = getSoundFX("GearDownWheel.wav", false);
			if (soundGearUpWheel == null || soundGearDownWheel == null) {
				gearSoundAvailable = false;
				wheelSoundAvailable = false;
			}
		}
		engineSoundAvailable = ((soundSpeedR = getSoundFX("SpeedR.wav", true)) != null &&
								(soundSpeedN = getSoundFX("SpeedN.wav", true)) != null &&
								(soundSpeed1 = getSoundFX("Speed1.wav", true)) != null &&
								(soundSpeed2 = getSoundFX("Speed2.wav", true)) != null &&
								(soundSpeed3 = getSoundFX("Speed3.wav", true)) != null);
	
		enableGearSound = gearSoundAvailable;
		enableWheelSound = wheelSoundAvailable;
		enableEngineSound = engineSoundAvailable;
		enableStatusSound = statusSoundAvailable;
	}
	
	/**
	 * Adds the menu commands for the sound FX.
	 */ 
	private void addCommands() {
		// Add commands
		if (enableWheelSound || enableGearSound || enableEngineSound || enableStatusSound) 
	        addCommand(soundOffCommand);
        setCommandListener(this);
	}
	
	/**
	 * Set positions for the Sprites.
	 */
	private void setSpritePos(boolean preferredLayout) {
		// The two different layouts (Preferred & not Preferred):
		// Not Preferred | Preferred layout
		//               |
		//  DDDDDDDDDD   |   SS DDDDDDDDDD
		//  DDDDDDDDDD   |   SS DDDDDDDDDD
		//               |   SS 
		//  SS PPPPPPP   |   SS PPPPPPPPPP
		//  SS           |   SS 
		//  SS WWWWWWW   |   SS WWWWWWWWWW
		//  SS WWWWWWW   |   SS WWWWWWWWWW
		//
		// (D=Display, P=Panel, S=Speed slider, W=Wheel)
		int screenWidth;
		int screenHeight;
		int dW, dH, dX, dY;
		int pW, pH, pX, pY;
		int wW, wH, wX, wY;
		int sW, sH, sX, sY;
		if (landscapeMode) {
			dH = getSprite(DISPLAY_IDLE).getWidth();
			dW = getSprite(DISPLAY_IDLE).getHeight();
			pH = getSprite(carAPI.CAR_READY).getWidth();
			pW = getSprite(carAPI.CAR_READY).getHeight();
			wH = getSprite(carAPI.CAR_DIRECTION_LEFT).getWidth();
			wW = getSprite(carAPI.CAR_DIRECTION_LEFT).getHeight();
			sH = getSprite(carAPI.CAR_SPEED_R).getWidth();
			sW = getSprite(carAPI.CAR_SPEED_R).getHeight();
			screenWidth = getHeight();
			screenHeight = getWidth();
		} else {
			dW = getSprite(DISPLAY_IDLE).getWidth();
			dH = getSprite(DISPLAY_IDLE).getHeight();
			pW = getSprite(carAPI.CAR_READY).getWidth();
			pH = getSprite(carAPI.CAR_READY).getHeight();
			wW = getSprite(carAPI.CAR_DIRECTION_LEFT).getWidth();
			wH = getSprite(carAPI.CAR_DIRECTION_LEFT).getHeight();
			sW = getSprite(carAPI.CAR_SPEED_R).getWidth();
			sH = getSprite(carAPI.CAR_SPEED_R).getHeight();
			screenWidth = getWidth();
			screenHeight = getHeight();
		}
		
		// Define position for name display
		if (preferredLayout) {
			dX = screenWidth - dW;
			dY = 0;
		} else {
			dX = (screenWidth - dW) / 2;
			dY = 0;
		}
		// Define positions for panel pictures
		if (preferredLayout) {
			pX = (screenWidth - dW) + ((dW - pW) / 2);
			pY = dH;
		} else {
			pX = screenWidth - pW;
			pY = dH;
		}
		// Define positions for speed pictures
		if (preferredLayout) {
			sX = (screenWidth - dW - sW) / 2;
			sY = screenHeight - sH;
		} else {
			sX = (screenWidth - pW - sW) / 2;
			sY = screenHeight - sH;
		}
		// Define positions for wheel pictures
		if (preferredLayout) {
			wX = (screenWidth - dW) + ((dW - wW) / 2);
			wY = screenHeight - wH;
		} else {
			wX = screenWidth - wW;
			wY = screenHeight - wH;
		}
		try {			
			Sprite tempSprite;
			// Set position for mode display
			for (int i = DISPLAY_PLAY; i <= DISPLAY_IDLE; i++) {
				tempSprite = getSprite(i);
				if (landscapeMode)
					tempSprite.setPosition(dY, screenWidth - dX - dW);
				else
					tempSprite.setPosition(dX, dY);
			}
			// Set positions for panel pictures
			for (int i = carAPI.CAR_READY; i <= carAPI.CAR_NO_LINK; i++) {
				tempSprite = getSprite(i);
				if (landscapeMode)
					tempSprite.setPosition(pY, screenWidth - pX - pW);
				else
					tempSprite.setPosition(pX, pY);
			}
			// Set positions for speed pictures
			for (int i = carAPI.CAR_SPEED_R; i <= carAPI.CAR_SPEED_3; i++) {
				tempSprite = getSprite(i);
				if (landscapeMode)
					tempSprite.setPosition(sY, screenWidth - sX - sW);
				else
					tempSprite.setPosition(sX, sY);
			}
			// Set positions for wheel pictures
			for (int i = carAPI.CAR_DIRECTION_LEFT; i <= carAPI.CAR_DIRECTION_RIGHT; i++) {
				tempSprite = getSprite(i);
				if (landscapeMode)
					tempSprite.setPosition(wY, screenWidth - wX - wW);
				else
					tempSprite.setPosition(wX, wY);
			}
		}
		catch (NullPointerException e) {
            System.out.println("NullPointerException: Set sprites position: " + 
            					e.getMessage());
		}
	}
		
	/**
	 * Insert <code>Sprite s</code> into the vector <code>sprites</code> on the
	 * position <code>pos+1</code>. This is because <code>pos</code> can be
	 * <code>-1</code>. 
	 *
	 * @param pos the position taken from <code>CarAPI</code>
	 * @param s the <code>Sprite</code> to be saved
	 */ 
	private void insertSprite(int pos, Sprite s) {
		pos++;
		sprites[pos] = s;
	}
	
	/**
	 * Converts an RGB triplet to an int. Used by downscaleImage() below.
	 *
	 * @param r	The red component
	 * @param g	The green component
	 * @param b	The blue component
	 * @return	The int representing the RGB color.
	 */
	private int rgbToInt(int r, int g, int b) {
		return (r << 16) | (g << 8) | (b);
	}
	
	/*
	 * Downscales the <code>sourceImage</code> to a percentage <code>p</code>.
	 * Fast and small but without filtering, so the result is not the best but
	 * good enough for most applications.
	 *
	 * @param p	The percentage to which to downscale the image.
	 * @param sourceImage	The image to downscale.
	 * @return	The downscaled image.
	 */
	private Image downscaleImage(int p, Image sourceImage) {
		// First some precautions:
		if (p >= 100) return Image.createImage(sourceImage);
		if (p <= 0) return null;
		// Set up some variables with values for the source image:
		int sourceWidth = sourceImage.getWidth();
		int sourceHeight = sourceImage.getHeight();
		int[] rgbDataSource = new int[sourceWidth * sourceHeight];
		// ...and calculate the target image size:
		int targetWidth = (sourceWidth * p) / 100;
		int targetHeight = (sourceHeight * p) / 100;
		int[] rgbDataTarget = new int[targetWidth * targetHeight];
		// Fill the rgbDataSource[] array with image data from the source image.
		// The data is stored as 4 bytes (= 32 bits, which fits nicely into an int)
		// whith the first byte representing transparency and the next three bytes
		// representing red, green, & blue.
		sourceImage.getRGB(rgbDataSource, 0, sourceWidth, 0, 0, sourceWidth, sourceHeight);
		// Now loop through all pixels in the target image, and for each pixel
		// calculate an average of the corresponding pixel values that this pixel
		// covers in the source image. No pixels in the source image are calculated
		// twice; partly covered pixels are either included or excluded (depending
		// on position).
		for (int yTarget = 0; yTarget < targetHeight; yTarget++) {
			for (int xTarget = 0; xTarget < targetWidth; xTarget++) {
				// Now we've picked a target pixel. Let's loop through all pixels
				// in the source image that's covered by this pixel.
				// But first reset our pixel buffer that is used for calculating
				// the avarage of covered pixels:
				int pxR = 0;
				int pxG = 0;
				int pxB = 0;
				int pxSampled = 0;
				for (int ySource = (100 * yTarget) / p; ySource < (100 * (yTarget + 1)) / p; ySource++) {
					for (int xSource = (100 * xTarget) / p; xSource < (100 * (xTarget + 1)) / p; xSource++) {
						// Make sure that the target pixel actually is inside the
						// source image (shouldn't be necessary...):
						if (xSource < sourceWidth) { 
							// Add the values of the red, green, & blue part of the
							// source pixel to our pixel buffer:
							int rgb = rgbDataSource[(ySource * sourceWidth) + xSource];
							pxR += (rgb >>> 16) & 0xFF;
							pxG += (rgb >>> 8) & 0xFF;
							pxB += rgb & 0xFF;
						}
						// Keep track of the number of pixels we've sampled:
						pxSampled++;
					} // End of for (int xSource...
				} // End of for (int ySource...
				// Now we've added all pixels in the source image that's covered
				// by this target pixel, so let's divide it by the nunber of sampled
				// pixels. (Since pxSampled can never be zero as long as p < 100,
				// there is no risk of divide by zero...):
				pxR = pxR / pxSampled;
				pxG = pxG / pxSampled;
				pxB = pxB / pxSampled;
				// Now add this averaged pixel in the target image array.
				rgbDataTarget[(yTarget * targetWidth) + xTarget] = rgbToInt(pxR, pxG, pxB);
			} // End of for (int xTarget...
		} // End of for (int yTarget...
		// Create an image out of the target image array and return it.
		return Image.createRGBImage(rgbDataTarget, targetWidth, targetHeight, false);
	} // End of downscaleImage()

	/**
	 * Get the <code>Sprite</code> in the vector <code>sprites</code> on 
	 * position <code>pos</code>.
	 *
	 * @param pos the position taken from <code>CarAPI</code>
	 * @return the <code>Sprite</code> on position <code>pos</code>
	 */
	private Sprite getSprite(int pos) {
		Sprite s;
		if (pos >= carAPI.CAR_SPEED_R && pos <= carAPI.CAR_SPEED_3) {
			s = sprites[pos+1];
			s.defineReferencePixel(s.getWidth() / 2, s.getWidth() / 2);
			if (!landscapeMode)
				s.setTransform(Sprite.TRANS_ROT90);
		}
		else {
			if (pos  >= DISPLAY_PLAY && pos <= DISPLAY_IDLE) {
				if (pos == DISPLAY_IDLE)
					s = spDisplayIdle;
				else if (pos == DISPLAY_REC)
					s = spDisplayRec;
				else 
					s = spDisplayPlay;
			}
			else
				s = sprites[pos+1];
			s.defineReferencePixel(s.getWidth() / 2, s.getWidth() / 2);
			if (landscapeMode)
				s.setTransform(Sprite.TRANS_ROT270);
			else
				s.setTransform(Sprite.TRANS_NONE);
		}
		return s;	
	}
	
	/**
	 * Called from the run thread when a sound fx is to be played.
	 * Any sound fx already playing will be stopped (paused) before the new
	 * sound is played, since only one sound can be played at a time.
	 * To somewhat try to compensate for this shortcomming, there are two composite sounds
	 * (soundGearUpWheel & soundGearDownWheel), which are the wheel sound and the gear sounds
	 * mixed together. These composite sounds are played when we want to play the wheel
	 * sound and one of the gear sounds at the same time.
	 *
     * @param player	The <code>Player</code> which shall be played.
	 */
	private void playSounds() {

		if (!wantToPlayWheel && !wantToPlayGearUp && !wantToPlayGearDown && 
			!wantToPlayReady && !wantToPlayCharging && !wantToPlayNoLink) {
			boolean playEngine = false;
			try {
				if (currentlyPlayingSound == null) 
					playEngine = true;
				else
				if (currentlyPlayingSound.getState() != Player.STARTED) 
					playEngine = true;
				if (playEngine) {
					switch (carAPI.getSpeed()) {
						case -1:
							currentlyPlayingSound = soundSpeedR;
							break;
						case 0:
							currentlyPlayingSound = soundSpeedN;
							break;
						case 1:
							currentlyPlayingSound = soundSpeed1;
							break;
						case 2:
							currentlyPlayingSound = soundSpeed2;
							break;
						case 3:
							currentlyPlayingSound = soundSpeed3;
							break;
					}
//					currentlyPlayingSound.setMediaTime(0);
					currentlyPlayingSound.start();
				}
			}
			catch(Exception e){}// {showMessage("Exception: " + e.toString());}
			
			return;
		}
		Player soundToPlay = null;

		if (wantToPlayWheel)
			startedWheelSoundTime = System.currentTimeMillis();
		else
			if (System.currentTimeMillis() - startedWheelSoundTime < soundWheel.getDuration() / 1000) 
				wantToPlayWheel = true;

		if (wantToPlayReady) {
			if (enableStatusSound)
				soundToPlay = soundReady;
			wantToPlayReady = false;
		} else
		if (wantToPlayCharging) {
			if (enableStatusSound)
				soundToPlay = soundCharging;
			wantToPlayCharging = false;
		} else
		if (wantToPlayNoLink) {
			if (enableStatusSound)
				soundToPlay = soundNoLink;
			wantToPlayNoLink = false;
		} 
		else {
			if (wantToPlayGearUp && wantToPlayWheel && enableWheelSound && enableGearSound) {
					soundToPlay = soundGearUpWheel;
			} else
			if (wantToPlayGearDown && wantToPlayWheel && enableWheelSound && enableGearSound) {
					soundToPlay = soundGearDownWheel;
			} else
			if (wantToPlayGearUp && enableGearSound) {
					soundToPlay = soundGearUp;
			} else
			if (wantToPlayGearDown && enableGearSound) {
					soundToPlay = soundGearDown;
			} else
			if (wantToPlayWheel && enableWheelSound) {
					soundToPlay = soundWheel;
			}
			wantToPlayWheel = false;
			wantToPlayGearUp = false;
			wantToPlayGearDown = false;
		}

		// Now stop (=pause) the currently playing sound, if any, and play the soundToPlay (if any).
		if (soundToPlay == null) return;
		try {
			if (currentlyPlayingSound != null) {
				currentlyPlayingSound.stop();
			}
			
			currentlyPlayingSound = soundToPlay;
			currentlyPlayingSound.setMediaTime(0);
			currentlyPlayingSound.start();
		}
		catch(Exception e){}// {showMessage("Exception: " + e.toString());}
	}	
	
	/**
	 * Returns the current mode, i.e. Idle = -2, Recording = -3, Replaying = -4.
	 *
	 * @return the current mode.
	 */
	 private int getMode() {
	 	if (recording) 
	 		return -3;
	 	else if (replaying)
	 		return -4;
	 	else return -2;
	 }
	 
	/**
	 * Processes any pending car events and repaint the canvas as needed.
	 * This is also where the sound FX are triggered. The sounds are played
	 * in <code>playSounds()</code>, which is called from the game loop (the
	 * <code>run()</code> thread).The <code>playSounds()</code> plays the FX
	 * sounds based on some flags (boolean starting with "wantToPlay..."), which
	 * are set here and cleared in <code>playSounds()</code> as they are being played.
	 *
	 * @param s the Sprite specifying the repaint area
	 */
	private void processCarEvents() {

		// The Display:
		if (lastKnownMode != getMode()) {
	    	repaint(getSprite(DISPLAY_IDLE).getX(), 
	    			getSprite(DISPLAY_IDLE).getY(), 
	    			getSprite(DISPLAY_IDLE).getWidth(), 
	    			getSprite(DISPLAY_IDLE).getHeight());
		}
		// The Wheel:
		if (carAPI.getDirection() != lastKnownDirection) {
			if (carAPI.getDirection() == carAPI.CAR_DIRECTION_LEFT || 
				carAPI.getDirection() == carAPI.CAR_DIRECTION_RIGHT)
			    wantToPlayWheel = true;
			if (recording) recordEvent(carAPI.getDirection());
	    	repaint(getSprite(carAPI.CAR_DIRECTION_LEFT).getX(), 
	    			getSprite(carAPI.CAR_DIRECTION_LEFT).getY(), 
	    			getSprite(carAPI.CAR_DIRECTION_LEFT).getWidth(), 
	    			getSprite(carAPI.CAR_DIRECTION_LEFT).getHeight());
		} 
		// The Speed:
		if (carAPI.getSpeed() != lastKnownSpeed) {
	    	if (carAPI.getSpeed() < lastKnownSpeed) 
				wantToPlayGearDown = true;
	    	else
				wantToPlayGearUp = true;
			if (recording) recordEvent(carAPI.getSpeed());
	    	repaint(getSprite(carAPI.CAR_SPEED_N).getX(), 
	    			getSprite(carAPI.CAR_SPEED_N).getY(), 
	    			getSprite(carAPI.CAR_SPEED_N).getWidth(), 
	    			getSprite(carAPI.CAR_SPEED_N).getHeight());
		}
		// The Status panel:
		if (carAPI.getStatus() != lastKnownStatus) {
			if (carAPI.getStatus() == carAPI.CAR_READY) {
				wantToPlayReady = true;
				carAPI.setSpeed(carAPI.CAR_SPEED_N);
				carAPI.setDirection(carAPI.CAR_DIRECTION_STRAIGHT);
				repaint();
			}
			else
			if (carAPI.getStatus() == carAPI.CAR_CHARGING)
				wantToPlayCharging = true;
			else
			if (carAPI.getStatus() == carAPI.CAR_NO_LINK)
				wantToPlayNoLink = true;
	    	repaint(getSprite(carAPI.CAR_READY).getX(), 
	    			getSprite(carAPI.CAR_READY).getY(), 
	    			getSprite(carAPI.CAR_READY).getWidth(), 
	    			getSprite(carAPI.CAR_READY).getHeight());
		} 

    	lastKnownSpeed = carAPI.getSpeed();
    	lastKnownDirection = carAPI.getDirection();
    	lastKnownStatus = carAPI.getStatus();
    	lastKnownMode = getMode();
	}

	/**
	 * Check if the <code>Sprite</code> intersects or is within the clip area.
	 *
	 * @param sp the <code>Sprite</code> to be tested.
	 * @param clipX the x-coordinate of the upper left corner of the clip area
	 * @param clipY the y-coordinate of the upper left corner of the clip area
	 * @param clipWidth the width of the clip area
	 * @param clipHeight the height of the clip area
	 * @return 	<code>true</code> if the <code>Sprite</code> intersects or is
	 *			within the clip area, otherwise <code>false</code>.  
	 */
	private boolean intersects(Sprite sp, int clipX, int clipY, 
							   int clipWidth, int clipHeight) {
					
		// Sprite is to the left of the clip area
		if (sp.getX()+sp.getWidth() < clipX)
			return false;
			
		// Sprite is to the right of the clip area
		if (sp.getX() > clipX+clipWidth)
			return false;
		
		// Sprite is above the clip area
		if (sp.getY()+sp.getHeight() < clipY) 
			return false;
			
		// Sprite is below the clip area
		if (sp.getY() > clipY+clipHeight)
			return false;
		
		// If Sprite is not outside of clip area it must intersect or be within
		// the clip are.
		return true;
	}
	
  	/**
	 * Check if the specified coordinates is within the <code>Sprite</code> 
	 * specified.
	 *
	 * @param sp the <code>Sprite</code> to be tested.
	 * @param x the horizontal location.
	 * @param y the vertical location.
	 */
	private boolean withinSprite(Sprite sp, int x, int y) {
		return (sp.getX() <= x && x <= sp.getX()+sp.getWidth() &&
				sp.getY() <= y && y <= sp.getY()+sp.getHeight());
	}
  
  	/**
  	 * Return the gear number depending on the x & y-coordinate.
  	 *
  	 * @param x the x-coordinate
  	 * @param y the y-coordinate
  	 * @return the gear number
  	 */
  	private int getGear(int x, int y) {  
  		int h;
  		if (landscapeMode) {
  			h = getSprite(carAPI.CAR_SPEED_N).getWidth();
			y = x - getSprite(carAPI.CAR_SPEED_N).getX();
  		} else {
	  		h = getSprite(carAPI.CAR_SPEED_N).getHeight();
			y = y - getSprite(carAPI.CAR_SPEED_N).getY();
  		}
  		if (y < (32 * h) / 208)
			return carAPI.CAR_SPEED_3;
		else if (y < (67 * h) / 208)
			return carAPI.CAR_SPEED_2;
		else if (y < (100 * h) / 208)
			return carAPI.CAR_SPEED_1;
		else if (y < (153 * h) / 208)
			return carAPI.CAR_SPEED_N;
		else
			return carAPI.CAR_SPEED_R;
  	}
  
  	/**
  	 * Return the direction depending on the x & y-coordinate.
  	 *
  	 * @param x the x-coordinate
  	 * @param y the y-coordinate
  	 * @return the direction
  	 */
  	private int getDirection(int x, int y) {  			
  		int w;
  		if (landscapeMode) {
  			w = getSprite(carAPI.CAR_DIRECTION_LEFT).getHeight();
			x = w - (y - getSprite(carAPI.CAR_DIRECTION_LEFT).getY());
  		} else {
	  		w = getSprite(carAPI.CAR_DIRECTION_LEFT).getWidth();
			x = x - getSprite(carAPI.CAR_DIRECTION_LEFT).getX();
  		}
  		if (x < (17 * w) / 100)
			return carAPI.CAR_DIRECTION_LEFT;
		else if (x < (33 * w) / 100)
			return carAPI.CAR_DIRECTION_LIGHT_LEFT;
		else if (x < (67 * w) / 100)
			return carAPI.CAR_DIRECTION_STRAIGHT;
		else if (x < (83 * w) / 100)
			return carAPI.CAR_DIRECTION_LIGHT_RIGHT;
		else
			return carAPI.CAR_DIRECTION_RIGHT;
  	}
  
  	/**
   	 * Invoked by system to draw canvas.
   	 *
   	 * @param g - the <code>Graphics</code> object to be rendered.
   	 */
  	public void paint(Graphics g) {
  			
		// Make sure that g != null
		if (g == null) {
			return;
		}
			
		int	clipX =	g.getClipX();
		int	clipY =	g.getClipY();
		int	clipWidth =	g.getClipWidth();
		int	clipHeight = g.getClipHeight();
	
		// Set background color.
		g.setColor(130, 158, 160);
		g.fillRect(clipX, clipY, clipWidth, clipHeight);
		if (blackScreen) {
			blackScreen = false;
			g.setColor(0, 0, 0);
			g.drawString(blackScreenText, getWidth() / 2, getHeight() / 2, Graphics.BOTTOM|Graphics.HCENTER);
		}
		
 		// Draw display
		if (intersects(getSprite(DISPLAY_IDLE), clipX, clipY, clipWidth, clipHeight)) {
			getSprite(getMode()).paint(g);
		}
	
		// Draw wheel pictures
		if (intersects(getSprite(carAPI.CAR_DIRECTION_LEFT), 
					   clipX, clipY, clipWidth, clipHeight)) {
		    switch (carAPI.getDirection()) {
		    case 7:
		    	getSprite(carAPI.CAR_DIRECTION_LEFT).paint(g);
		    	break;
		    case 8:
		    	getSprite(carAPI.CAR_DIRECTION_LIGHT_LEFT).paint(g);
		    	break;
		    case 9:
		    	getSprite(carAPI.CAR_DIRECTION_STRAIGHT).paint(g);
		    	break;
		    case 10:
		    	getSprite(carAPI.CAR_DIRECTION_LIGHT_RIGHT).paint(g);
		    	break;
		    case 11:
		    	getSprite(carAPI.CAR_DIRECTION_RIGHT).paint(g);
		    	break;
		  	}
		}
		
		// Draw panel pictures
		if (intersects(getSprite(carAPI.CAR_READY), 
					   clipX, clipY, clipWidth, clipHeight)) {
			switch (carAPI.getStatus()) {
		    case 4:
		    	getSprite(carAPI.CAR_READY).paint(g);
		    	break;
		    case 5:
		    	getSprite(carAPI.CAR_CHARGING).paint(g);
		    	break;
		    case 6:
		    	getSprite(carAPI.CAR_NO_LINK).paint(g);
		    	break;
		  	}
		}

		// Draw speed pictures.
		if (intersects(getSprite(carAPI.CAR_SPEED_R), 
					   clipX, clipY, clipWidth, clipHeight)) {
		   	switch (carAPI.getSpeed()) {
		    case -1:
		    	getSprite(carAPI.CAR_SPEED_R).paint(g);
		    	break;
		    case 0:
		    	getSprite(carAPI.CAR_SPEED_N).paint(g);
		    	break;
		    case 1:
		    	getSprite(carAPI.CAR_SPEED_1).paint(g);
		    	break;
		    case 2:
		    	getSprite(carAPI.CAR_SPEED_2).paint(g);
		    	break;
		    case 3:
		    	getSprite(carAPI.CAR_SPEED_3).paint(g);
		    	break;
		  	}
		}
	
  	}

	/**
	 * Called when the drawable area of the Canvas has been changed.
	 * Calles the loadImages() method if the width or height is different from
	 * what it was the last time loadImages() was executed. loadImages() will
	 * load all images if not already done and downscale any images as needed.
	 *
	 * @param w the new width in pixels of the drawable area of the 
	 *			<code>Canvas</code>
	 * @param h the new height in pixels of the drawable area of the 
	 *			<code>Canvas</code>
	 */
	protected void sizeChanged(int w, int h) {		
		if (h > w)
			landscapeMode = true;
		else
			landscapeMode = false;
		if (!started) return;
		if (w != lastKnownWidth || h != lastKnownHeight) {
			blackScreen = true;
			blackScreenText = "Adjusting graphics...";
			repaint();
			serviceRepaints();
			loadImages();
		}
	}

	/**
	 * Translates game actions (RIGHT, LEFT, UP, & DOWN) from normal screen-mode
	 * to landscape-mode.
	 */
	 private int getLandscapeGameAction(int code) {
	 	switch (code) {
		 	case RIGHT:	return DOWN;
		 	case LEFT:	return UP;
		 	case UP:	return RIGHT;
		 	case DOWN:	return LEFT;
	 	}
	 	return code;
	 }
  	/**
  	 * Handles user input from the buttons.
  	 *
  	 * @param keyCode the code for the key action generated by user.
  	 */
  	protected void keyPressed(int keyCode) {
  		
  		//System.out.println("keyCode:" + keyCode);
  		if (replaying) stopReplaying();
  		// Translate to game actions.
  		int code;
		if (keyCode == P900_OK) code = LEFT;
		else if (keyCode == P900_MENU) code = RIGHT;
		else code = getGameAction(keyCode);
		if (landscapeMode)
			code = getLandscapeGameAction(code);
		// Handle game actions.
		switch (code) {
		case DOWN:
			if (carAPI.getSpeed() > carAPI.CAR_SPEED_N) {
				//System.out.println("Stop car");
				carAPI.setSpeed(carAPI.CAR_SPEED_N);
			}
			else {
				//System.out.println("Drive backward");
				carAPI.setSpeed(carAPI.CAR_SPEED_R);
			}	
			// Repaint the gear.
			processCarEvents();
		  	break;
		case UP:
			//System.out.println("Increase speed");
			carAPI.incSpeed();
			// Repaint the gear.
			processCarEvents();
		  	break;
		case RIGHT:
			//System.out.println("Turn right");
			carAPI.turnRight();
			// Repaint the steering wheel.
			processCarEvents();
		  	break;
		case LEFT:
			//System.out.println("Turn left");
			carAPI.turnLeft();
			// Repaint the steering wheel.
			processCarEvents();
		  	break;
		case FIRE:
			// Do nothing
		  	break;
		}		
	}

  	/**
  	 * Handles user input from the buttons.
  	 *
  	 * @param keyCode the code for the key action generated by user.
  	 */
  	protected void keyRepeated(int keyCode) {
  		
  		//System.out.println("keyCode:" + keyCode);
  		if (replaying) stopReplaying();
  		// Translate to game actions.
  		int code;
		if (keyCode == P900_OK) code = LEFT;
		else if (keyCode == P900_MENU) code = RIGHT;
		else code = getGameAction(keyCode);
		if (landscapeMode)
			code = getLandscapeGameAction(code);
		// Handle game actions.
		switch (code) {
		case DOWN:
			//System.out.println("Decrease speed");
			carAPI.decSpeed();
			// Repaint the gear.
			processCarEvents();
		  	break;
		case UP:
			//System.out.println("Increase speed");
			carAPI.incSpeed();
			// Repaint the gear.
			processCarEvents();
		  	break;
		case RIGHT:
			//System.out.println("Turn right");
			carAPI.turnRight();
			// Repaint the steering wheel.
			processCarEvents();
		  	break;
		case LEFT:
			//System.out.println("Turn left");
			carAPI.turnLeft();
			// Repaint the steering wheel.
			processCarEvents();
		  	break;
		}
		
	}

  	/**
  	 * Handles user input from the buttons.
  	 *
  	 * @param keyCode the code for the key action generated by user.
  	 */
  	protected void keyReleased(int keyCode) {
  		
  		//System.out.println("keyCode:" + keyCode);
  		if (replaying) stopReplaying();
  		// Translate to game actions.
  		int code;
		if (keyCode == P900_OK) code = LEFT;
		else if (keyCode == P900_MENU) code = RIGHT;
		else code = getGameAction(keyCode);

		// Handle game actions.
		switch (code) {
		case RIGHT:
		  	break;
		case LEFT:
		  	break;
		case UP:
		  	break;
		case DOWN:
		  	break;
		case FIRE:
			//System.out.println("Go straight");
			carAPI.setDirection(carAPI.CAR_DIRECTION_STRAIGHT);
			processCarEvents();
		  	break;
		}
	}

	/**
	 * Called when the pointer is pressed. Changes speed and/or direction of the car.
	 * (direction is changed if the pointer is within the wheel sprite)
	 *
	 * @param x the horizontal location where the pointer was pressed (relative 
	 *			to the Canvas)
	 * @param y the vertical location where the pointer was pressed (relative 
	 *			to the Canvas)
	 */
	protected void pointerPressed(int x, int y) {
			
  		if (replaying) stopReplaying();
		int gear = getGear(x, y);
		carAPI.setSpeed(gear);
		processCarEvents();
		enableTouchWheel = withinSprite(getSprite(carAPI.CAR_DIRECTION_STRAIGHT), x, y);
		if (enableTouchWheel) {
			int direction = getDirection(x, y);
			carAPI.setDirection(direction);
			processCarEvents();
		}
			
	}

	/**
	 * Called when the pointer is dragged. Changes speed and/or direction of the car.
	 * (direction is changed if the pointer was previously pressed within the wheel sprite)
	 *
	 * @param x the horizontal location where the pointer was dragged (relative 
	 *			to the Canvas)
	 * @param y the vertical location where the pointer was dragged (relative 
	 *			to the Canvas)
	 */
	protected void pointerDragged(int x, int y) {
				
  		if (replaying) stopReplaying();
		int gear = getGear(x, y);
		carAPI.setSpeed(gear);
		processCarEvents();
		if (enableTouchWheel) {
			int direction = getDirection(x, y);
			carAPI.setDirection(direction);
			processCarEvents();
		}
	}
	
	/**
	 * Called by the system when the pointer is released. Changes speed and/or direction of the car.
	 * (direction is changed if the pointer was previously pressed within the wheel sprite)
	 *
	 * @param x the horizontal location where the pointer was released 
	 *			(relative to the Canvas)
	 * @param y the vertical location where the pointer was released 
	 *			(relative to the Canvas)
	 */
	protected void pointerReleased(int x, int y) {
				
  		if (replaying) stopReplaying();
		carAPI.setSpeed(carAPI.CAR_SPEED_N);
		processCarEvents();
		if (enableTouchWheel) {
			carAPI.setDirection(carAPI.CAR_DIRECTION_STRAIGHT);
			processCarEvents();
		}
	}
	
	/**
	 * Called by the system to notify us that the canvas is now at least
	 * partially obscured by another application window, when it previously has
	 * been completely visible.
	 * Calls the <code>carAPI.setAppActive(false)</code>.
	 */
	protected void hideNotify() {
		if (started)
			carAPI.setAppActive(false);
		if (recording) stopRecording();
		if (replaying) stopReplaying();
	}
	
	/**
	 * Called by the system to notify us that the canvas is now completely
	 * visible, when it previously has been at least partially obscured by
	 * another application window.
	 * Calls the <code>carAPI.setAppActive(true)</code>.
	 */
	protected void showNotify() {
		if (started)
			carAPI.setAppActive(true);
	}

	private void startRecording() {
		if (replaying) stopReplaying();
		drivePath.removeAllElements();
		eventStartTime = 0;
		recording = true;
		removeCommand(recordCommand);
		removeCommand(replayCommand);
		removeCommand(stopReplayCommand);
		addCommand(recordCommand);
		addCommand(stopRecordCommand);
	}

	public void stopRecording() {
		recording = false;
		// set the event duration of the previous recorded event (if any):
		if (eventStartTime > 0)
			try {
				driveEventClass driveEvent = (driveEventClass) drivePath.lastElement();
				driveEvent.eventDuration = System.currentTimeMillis() - eventStartTime;
			}
			catch (NoSuchElementException e) {}
		eventStartTime = 0;
		removeCommand(stopRecordCommand);
		removeCommand(stopReplayCommand);
		addCommand(recordCommand);
		if (!drivePath.isEmpty())
			addCommand(replayCommand);
	}

	private void startReplaying() {
		recording = false;
		if (replaying) stopReplaying();
		replaying = true;
		events = drivePath.elements();
		playbackTimer = new Timer();
		playbackTimer.schedule(playbackTimerTask, 1); 
		removeCommand(recordCommand);
		removeCommand(stopRecordCommand);
		addCommand(stopReplayCommand);
	}

	public void stopReplaying() {
		removeCommand(stopRecordCommand);
		removeCommand(stopReplayCommand);
		addCommand(recordCommand);
		replaying = false;
	    if (playbackTimer != null) {
	        playbackTimer.cancel();
	        playbackTimer = null;
	    }
	}

	private void recordEvent(int event) {
		// set the event duration of the previous recorded event (if any):
		if (eventStartTime > 0)
			try {
				driveEventClass driveEvent = (driveEventClass) drivePath.lastElement();
				driveEvent.eventDuration = System.currentTimeMillis() - eventStartTime;
			}
			catch (NoSuchElementException e) {}
		driveEventClass driveEvent = new driveEventClass();
		driveEvent.event = event;
		driveEvent.eventDuration = -1;
		drivePath.addElement(driveEvent);
		eventStartTime = System.currentTimeMillis();
	}

    /**
     * Respond to menu commands.
     *
     * @param c	the <code>command</code> triggering this action
     * @param d the <code>Displayable</code> on which this event has occurred
     */
    public void commandAction(Command c, Displayable d) {
    	
    	System.out.println("Command: " + c.getLabel());
    	
    	// Destroy this MIDlet
		if (c == exitCommand) {
		    BluetoothCar.getInstance().destroyApp(false);
		} else
		if (c == soundOffCommand) {
			removeCommand(soundOffCommand);
			addCommand(soundOnCommand);
			enableWheelSound = false;
			enableGearSound = false;
			enableEngineSound = false;
			enableStatusSound = false;
		} else
		if (c == soundOnCommand) {
			removeCommand(soundOnCommand);
			addCommand(soundOffCommand);
			enableWheelSound = wheelSoundAvailable;
			enableGearSound = gearSoundAvailable;
			enableEngineSound = engineSoundAvailable;
			enableStatusSound = statusSoundAvailable;
		} else
		if (c == aboutCommand) {
			showAbout();
		} else
		if (c == recordCommand) {
			startRecording();
		} else
		if (c == stopRecordCommand) {
			stopRecording();
		} else
		if (c == replayCommand) {
			startReplaying();
		} else
		if (c == stopReplayCommand) {
			stopReplaying();
		} else
		if (c == rotateScrCommand) {
			landscapeMode = !landscapeMode;
			blackScreen = true;
			blackScreenText = "Adjusting graphics...";
			repaint();
			serviceRepaints();
			loadImages();
		}		
    }
    
    /**
     * The About box. Displays an image of the Bluetooth Car and the version
     * together with a message -both defined in BluetoothCar.java.
     */
	private void showAbout() {
	    Alert about = new Alert("Bluetooth Car ver. " + bluetoothCar.versionString,
	    						bluetoothCar.aboutString, 
	    						aboutImage, 
	    						AlertType.INFO);
	    about.setTimeout(Alert.FOREVER);
	    Display.getDisplay(BluetoothCar.getInstance()).setCurrent(about);
	}

	private void showMessage(String message) {
	    Alert alert = new Alert("Message:", message, null, AlertType.INFO);
	    alert.setTimeout(Alert.FOREVER);
	    Display.getDisplay(BluetoothCar.getInstance()).setCurrent(alert);
	}
}
