/**
 * SPPServer.java 
 */

import java.io.*;
import javax.microedition.io.*;
import javax.bluetooth.*;

/**
 * Copyright (c) 2003 Sony Ericsson Mobile Communications AB<br>
 * All Rights Reserved.
 * <p>
 * SPPServer is a Serial Port Profile server. The server will define a Serial
 * Port Profile service and then wait for client to connect. When a client
 * connect to the server a connection to the car will be created and the 
 * <code>CarAPI</code> will be passed to the <code>CarConnection</code> object.
 *
 * @author	Christopher Wong
 * @author	ke Rydgren
 * @version	2.1
 */
public class SPPServer extends Thread {
	
	// The service.
	private StreamConnectionNotifier service = null;
	
	// The connection to the remote device.
	private StreamConnection con = null;
	
	// Used to control car and check car status.
	private CarAPI carAPI;

	// The CarConnection, created when the car connects.
	private CarConnection carConnection;	
	
	private BluetoothCar bluetoothCar;
	
	private InputStream is;
	private OutputStream os;
	
	private boolean running = true;
	
	/**
	 * Constructs a new <code>SPPServer</code> object, which will create the
	 * <code>CarConnection</code> object, define the service for the serial port
	 * (SPP) and then start running itself.
	 *
	 * @param carAPI 	the interface to control car and check car status. This
	 *					will be passed to the <code>CarConnection</code> object
	 *					when it is created.
	 */
    public SPPServer(CarAPI carAPI, BluetoothCar bluetoothCar) {
    	this.carAPI = carAPI;
    	this .bluetoothCar = bluetoothCar;
		// Create the connection interface to the car.
		carConnection = new CarConnection(carAPI);
		// define the service for the serial port (SPP).
		defineService();
    	// Create this thread.
    	start();
    }
    
    /**
     * This is called by the constructor of the class. It will wait for a client
     * to connect and then start looping the event handler that takes care of any
     * user events or AT-commands from the car.
     *
     * This thread will only terminate when the program exit, but as long as
     * there are no connection open, the only activity in this thread is the
     * blocking call to StreamConnectionNotifier.acceptAndOpen().
     */
	public void run() {
		while (running) {
			// Wait (blocking) for a device to connect and accept that request.
			waitForClientConnection();
			// Now a device have connected to us.
			if (con != null) { // Just to make sure the connection was successful...
				if (!carAPI.isAppActive()) {
					// OK, we have a valid connection, but unfortunately we are in
					// the paused state.
					
					// Tell the application management software that we want to 
					// become active (may work in some future implementation...)
					bluetoothCar.resumeRequest();
					// Start a timer that will close the connection if we still are
					// in the paused state after 10 seconds.
					carAPI.startDisconnectTimer();
				}
				// Open input and output streams.
				try {
					is = con.openInputStream();
					os = con.openOutputStream();
				} 
				catch (IOException e) {
		            System.out.println("IOException: " + e.getMessage());
					closeConnection();
				}
				// Loop through the processEvents() method, which processes all
				// messages from the car (AT-commands) and all user events from
				// the GUI (via the CarAPI), until the connection is closed or
				// the variable carAPI.timeToCloseConnection is set to false by
				// the timer event that is started when this MIDlet is paused or
				// not in the foreground.
				carConnection.resetCarData();
				carAPI.timeToCloseConnection = false;
				while (carConnection.processEvents(is, os)) {
					// Sleep for a short while to let other treads execute.
		      		try { sleep(10); } catch (InterruptedException e) {}
				}
			}
			closeConnection();
		} // while (true)
	}
	
	/**
	 * Closes the connection to the car.
	 */
	private void closeConnection() {
		try {
			carConnection.quitCarLink(os);
			if (con != null) {
				con.close(); // Won't close until is & os closes...
				con = null;
			}
			if (is != null) {
				is.close();
				is = null;
			}
			if (os != null) {
				os.close();
				os = null;
			}
		} 
		catch (IOException e) {}	
	}
	
	public boolean connectionOpen() {
		return con != null;
	}
	
	/**
	 * Closes the connection and the SPP service.
	 */
	 public void closeAll() {
	 	closeConnection();
		try {
			service.close();
		} 
		catch (IOException e) {}	
		running = false;
	 }
	
    /**
     * Define the service this server provides (SPP).
     */
    private void defineService() {
		String url = " ";
		try {
			// 0x1101 for spp
			UUID myService = new UUID("1101", true);
			url = "btspp://localhost:" + myService.toString() +
			      ";name=xSPPServer";
			System.out.println(url);
		}
		catch (NumberFormatException e) {
            System.out.println("NumberFormatException: " + e.getMessage());
		}
		catch (IllegalArgumentException e) {
            System.out.println("IllegalArgumentException: " + e.getMessage());
		}
		catch (NullPointerException e) {
            System.out.println("NullPointerException: " + e.getMessage());
		}
					
		try {
			// Assigns a RFCOMM server channel and creates a service record
			// using this channel.
			service = (StreamConnectionNotifier) Connector.open(url);
			System.out.println("Connector.open(serviceURL): OK!");
		}
		catch (IllegalArgumentException e) {
            System.out.println("IllegalArgumentException: " + e.getMessage());
		}
		catch (ConnectionNotFoundException e) {
            System.out.println("ConnectionNotFoundException: " + e.getMessage());
		}
		catch (IOException e) {
            System.out.println("IOException: " + e.getMessage());
        }
	}
	
	/**
	 * Waits for client to connect and accepts that connection.
	 * This method will lock the invoker.
	 *
	 * If there are any Exceptions, this function returns false and the
	 * connection con is closed.
	 */
	private void waitForClientConnection() {
		try {
			// Add the service record to the SDDB and accept a client 
			// connection.
			System.out.println("Waits for client to connect!");
			System.out.flush();
			con = (StreamConnection) service.acceptAndOpen();
			
		} 
		catch (IOException e) {
            System.out.println("IOException: " + e.getMessage());
            closeConnection();
		}
    }

}
